@@ -11,9 +11,9 @@
 #define BLOSC_FASTCOPY_H
 
 /* Same semantics than memcpy() */
-unsigned char *fastcopy(unsigned char *out, const unsigned char *from, unsigned len);
+unsigned char *blosc_internal_fastcopy(unsigned char *out, const unsigned char *from, unsigned len);
 
-/* Same as fastcopy() but without overwriting origin or destination when they overlap */
-unsigned char* safecopy(unsigned char *out, const unsigned char *from, unsigned len);
+/* Same as blosc_internal_fastcopy() but without overwriting origin or destination when they overlap */
+unsigned char* blosc_internal_copy_match(unsigned char *out, const unsigned char *from, unsigned len);
 
 #endif /*BLOSC_FASTCOPY_H*/
